(function (global, factory) {
	typeof exports === 'object' && typeof module !== 'undefined' ? factory(exports, require('serializr'), require('mobx')) :
	typeof define === 'function' && define.amd ? define(['exports', 'serializr', 'mobx'], factory) :
	(factory((global['mobx-decorated-models'] = global['mobx-decorated-models'] || {}),global.serializr,global.mobx));
}(this, (function (exports,serializr,mobx) { 'use strict';

function isSerializable(model, property) {
    return !!(!model.$nonSerializable || !model.$nonSerializable[property]);
}

function markNonserializable(model, property) {
    model.$nonSerializable = model.$nonSerializable || {};
    model.$nonSerializable[property] = true;
}

var _typeof = typeof Symbol === "function" && typeof Symbol.iterator === "symbol" ? function (obj) {
  return typeof obj;
} : function (obj) {
  return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
};























































var toConsumableArray = function (arr) {
  if (Array.isArray(arr)) {
    for (var i = 0, arr2 = Array(arr.length); i < arr.length; i++) arr2[i] = arr[i];

    return arr2;
  } else {
    return Array.from(arr);
  }
};

function hasInterceptors(model) {
    return Boolean(model.constructor.$mxdm && model.constructor.$mxdm.interceptors.length);
}

function getSchema(klass) {
    if (klass.$mxdm) {
        return klass.$mxdm;
    }

    Object.defineProperty(klass, '$mxdm', {
        enumerable: false,
        writable: false,
        configurable: true,
        value: Object.freeze({
            interceptors: [],
            properties: new Map()
        })
    });
    return klass.$mxdm;
}

function addLazyInitializer(target, fn) {
    var _getSchema = getSchema(target.constructor),
        interceptors = _getSchema.interceptors;

    interceptors.push(fn);
}

function attachInterceptors(model) {
    if (!hasInterceptors(model) || model.$mxdbDidAttachInterceptors) {
        return;
    }
    Object.defineProperty(model, '$mxdbDidAttachInterceptors', {
        enumerable: false, writable: false, configurable: true, value: true
    });
    getSchema(model.constructor).interceptors.forEach(function (fn) {
        return fn(model);
    });
}

function setupModel(_ref) {
    var attrs = _ref.attrs,
        Klass = _ref.modelClass,
        array = _ref.array,
        defaultAttributes = _ref.defaultAttributes,
        inverseOf = _ref.inverseOf,
        parentModel = _ref.parentModel,
        parentModelProp = _ref.parentModelProp;

    if (!attrs) {
        return attrs;
    }
    var isObject = 'object' === (typeof attrs === 'undefined' ? 'undefined' : _typeof(attrs));
    var canWrite = Object.isExtensible(attrs);

    if (isObject && canWrite) {
        if (defaultAttributes) {
            if ('function' === typeof defaultAttributes) {
                defaultAttributes = defaultAttributes.call(parentModel, array, parentModel);
            }
            Object.keys(defaultAttributes).forEach(function (key) {
                if (!attrs[key]) {
                    attrs[key] = defaultAttributes[key];
                }
            });
        }

        if (inverseOf) {
            attrs[inverseOf] = parentModel;
            if (parentModelProp) {
                attrs[inverseOf + '_association_name'] = parentModelProp;
            }
        }
    }

    var model = Klass && 'function' === typeof Klass && !(attrs instanceof Klass) ? new Klass(attrs) : attrs;

    if (inverseOf) {
        markNonserializable(model, inverseOf);
    }

    return model;
}

var ModelsMap = Object.create(null);

var defaultModelLookup = function defaultModelLookup(_ref) {
    var property = _ref.property,
        _ref$model = _ref.model,
        model = _ref$model === undefined ? property : _ref$model;

    if ('function' === typeof model) {
        return model;
    }
    return ModelsMap[model] || ModelsMap[property];
};
var modelLookup = defaultModelLookup;

var defaultRememberModel = function defaultRememberModel(model) {
    ModelsMap[model.identifiedBy || model.name] = model;
};
var rememberModel = defaultRememberModel;

function registerModel(model) {
    rememberModel(model);
}

function findModel(model, property) {
    return modelLookup({ model: model, property: property });
}

function rememberModelUsing(fn) {
    rememberModel = fn || defaultRememberModel;
}

function lookupModelUsing(fn) {
    modelLookup = fn || defaultModelLookup;
}

var PendingLookups = [];

function objectSerializer() {
    return {
        serializer: function serializer(obj, prop, parent) {
            return isSerializable(parent, prop) ? obj : serializr.SKIP;
        },
        deserializer: function deserializer(json, cb) {
            cb(null, json);
        }
    };
}

function addReference(parentModel, propName, options, cb) {
    var model = findModel(options.model, propName);
    // if (propName === 'boxes')
    //     console.log("Add ref", options); // && model.constructor);
    if (model) {
        serializr.getDefaultModelSchema(parentModel).props[propName] = cb(model, propName, options);
    } else {
        serializr.getDefaultModelSchema(parentModel).props[propName] = objectSerializer();
        PendingLookups.push({
            parentModel: parentModel, propName: propName, options: options, cb: cb
        });
    }
}

var dateSerializer = function dateSerializer() {
    return serializr.custom(function (date) {
        return date ? date.toJSON() : null;
    }, function (json) {
        return json ? new Date(json) : null;
    });
};

var CustomSerializerTypes = {
    date: dateSerializer,
    object: objectSerializer,
    array: serializr.list
};

function getSerializer(options, defaultSerializer) {
    if (options.serializer) {
        return serializr.custom(options.serializer[0], options.serializer[1]);
    }
    var fns = (CustomSerializerTypes[options.type] || defaultSerializer)();
    if (options.writeOnly) {
        return {
            serializer: function serializer() {
                return serializr.SKIP;
            },
            deserializer: fns.deserializer
        };
    }
    return fns;
}

var Fields = {
    identifier: function identifier$$1(options) {
        return getSerializer(options, serializr.identifier);
    },
    field: function field(options) {
        return getSerializer(options, serializr.primitive);
    },
    session: function session(options) {
        return getSerializer(Object.assign({}, options, { writeOnly: true }), serializr.primitive);
    }
};

function associationSerializers(modelClass, propName, options) {
    return {
        serializer: options.serializer || modelClass.serialize || serializr.object(serializr.getDefaultModelSchema(modelClass)).serializer,
        deserializer: options.deserializer || modelClass.deserializer || function (attrs, context) {
            var model = void 0;
            // double equals catches null/undefined
            if (null == attrs) {
                attrs = {};
            }
            if (attrs instanceof modelClass) {
                model = attrs;
            } else {
                model = setupModel(Object.assign({}, options, { attrs: attrs, modelClass: modelClass }));
            }
            if (options.inverseOf) {
                markNonserializable(model, options.inverseOf);
                model[options.inverseOf + '_association_name'] = propName;
                model[options.inverseOf] = context.target;
            }

            return model;
        }
    };
}

var Associations = {
    hasMany: function hasMany(modelClass, propName, options) {
        var _associationSerialize = associationSerializers(modelClass, propName, options),
            _deserializer = _associationSerialize.deserializer,
            _serializer = _associationSerialize.serializer;

        options.model = modelClass;
        return Object.assign({}, options, {
            deserializer: function deserializer(collection, cb, context) {
                var value = collection && collection.map ? collection.map(function (attrs) {
                    return _deserializer(attrs, context);
                }) : collection;
                cb(null, value);
            },
            serializer: function serializer(collection, name, parent) {
                if (!isSerializable(parent, name)) {
                    return serializr.SKIP;
                }
                return collection && collection.map ? collection.map(function (model) {
                    return _serializer(model, name, parent);
                }) : collection;
            }
        });
    },
    belongsTo: function belongsTo(modelClass, propName, options) {
        var _associationSerialize2 = associationSerializers(modelClass, propName, options),
            _deserializer2 = _associationSerialize2.deserializer,
            _serializer2 = _associationSerialize2.serializer;

        options.model = modelClass;
        return Object.assign({}, options, {
            deserializer: function deserializer(attrs, cb, context) {
                cb(null, _deserializer2(attrs, context));
            },
            serializer: function serializer(belongsTo, name, parent) {
                if (!isSerializable(parent, name)) {
                    return serializr.SKIP;
                }
                return _serializer2(belongsTo, name, parent);
            }
        });
    }
};

var MixedInInstanceMethods = {
    serialize: function serialize$$1() {
        var schema = serializr.getDefaultModelSchema(this.constructor);
        return serializr.serialize(schema, this);
    },
    update: function update$$1(json, callback) {
        return serializr.update(serializr.getDefaultModelSchema(this.constructor), this, json, callback);
    }
};

var MixedInClassMethods = {
    deserialize: function deserialize$$1(json, callback) {
        return serializr.deserialize(serializr.getDefaultModelSchema(this), json, callback);
    }
};

function decorateModel(model) {
    Object.assign(model.prototype, MixedInInstanceMethods);
    Object.assign(model, MixedInClassMethods);

    registerModel(model);
    var serializeSchema = {};

    var _getSchema = getSchema(model),
        properties = _getSchema.properties;

    properties.forEach(function (_ref, name) {
        var type = _ref.type,
            options = _ref.options;

        if (Fields[type]) {
            serializeSchema[name] = Fields[type](options, name);
        }
    });

    serializr.createModelSchema(model, serializeSchema, function (_ref2) {
        var json = _ref2.json;

        // eslint-disable-next-line new-cap
        var instance = new model(json);
        attachInterceptors(instance);
        return instance;
    });
    properties.forEach(function (_ref3, name) {
        var type = _ref3.type,
            options = _ref3.options;

        if (Associations[type] || options.model) {
            addReference(model, name, options, Associations[type] || Associations.belongsTo);
        }
    });
    for (var i = PendingLookups.length - 1; i >= 0; i -= 1) {
        var _PendingLookups$i = PendingLookups[i],
            parentModel = _PendingLookups$i.parentModel,
            propName = _PendingLookups$i.propName,
            options = _PendingLookups$i.options,
            cb = _PendingLookups$i.cb;

        var referencedModel = findModel(options.model, propName);
        if (referencedModel) {
            var parentModelSchema = serializr.getDefaultModelSchema(parentModel);
            parentModelSchema.props[propName] = cb(referencedModel, propName, options);
            PendingLookups.splice(i, 1);
        }
    }
}

function registerCustomType(id, serializer) {
    CustomSerializerTypes[id] = function () {
        return serializr.custom(serializer.serialize, serializer.deserialize);
    };
}

function identifiedBy(modelId) {
    return function (model) {
        Object.defineProperty(model, 'identifiedBy', { value: modelId, writable: false });
        return decorateModel(model);
    };
}

function unresolvedAssociations() {
    return PendingLookups.map(function (_ref4) {
        var model = _ref4.parentModel,
            property = _ref4.propName;
        return { model: model, property: property };
    });
}

function onCollectionChangeInterceptor(_ref, parentModel, parentModelProp) {
    var modelClass = _ref.modelClass,
        model = _ref.model,
        defaultAttributes = _ref.defaults,
        inverseOf = _ref.inverseOf;

    return function (change) {
        if (!change.newValue) {
            change.newValue = {};
        }
        if (!modelClass) {
            modelClass = findModel(model, parentModelProp);
        }
        // console.log("ARY CHANGE");
        if (change.type === 'splice') {
            for (var i = 0; i < change.added.length; i += 1) {
                change.added[i] = setupModel({
                    attrs: change.added[i],
                    array: change.object,
                    modelClass: modelClass,
                    defaultAttributes: defaultAttributes,
                    inverseOf: inverseOf,
                    parentModel: parentModel,
                    parentModelProp: parentModelProp });
            }
        } else if (change.type === 'update') {
            change.newValue = setupModel({
                attrs: change.newValue,
                array: change.object,
                modelClass: modelClass,
                defaultAttributes: defaultAttributes,
                inverseOf: inverseOf,
                parentModel: parentModel,
                parentModelProp: parentModelProp });
        }
        return change;
    };
}

function createCollection(options, parentModel, parentModelProp) {
    var ary = mobx.observable.array([]);
    //console.log("CREATE COL", parentModelProp);
    if (options) {
        if (options.model) {
            ary.intercept(onCollectionChangeInterceptor(options, parentModel, parentModelProp));
        }
        if (options.extend) {
            if ('function' === typeof options.extend) {
                options.extend(ary);
            } else {
                Object.assign(ary, options.extend);
            }
        }
    }

    return ary;
}

function onPropertySet(change, _ref) {
    var modelClass = _ref.modelClass,
        defaultAttributes = _ref.defaultAttributes,
        inverseOf = _ref.inverseOf,
        parentModel = _ref.parentModel,
        parentModelProp = _ref.parentModelProp;

    change.newValue = setupModel({
        attrs: change.newValue,
        modelClass: modelClass,
        defaultAttributes: defaultAttributes,
        inverseOf: inverseOf,
        parentModel: parentModel,
        parentModelProp: parentModelProp
    });
    return change;
}

function onHasManySet(change, options) {
    if (change.type !== 'update' || !change.newValue) {
        return change;
    }
    var modelClass = options.modelClass,
        defaultAttributes = options.defaultAttributes,
        inverseOf = options.inverseOf,
        parentModel = options.parentModel,
        parentModelProp = options.parentModelProp;


    var array = mobx.observable.array([]);
    for (var i = 0; i < change.newValue.length; i += 1) {
        array.push(setupModel({
            attrs: change.newValue[i],
            array: change.newValue,
            modelClass: modelClass,
            defaultAttributes: defaultAttributes,
            inverseOf: inverseOf,
            parentModel: parentModel,
            parentModelProp: parentModelProp
        }));
    }
    if (parentModel[parentModelProp]) {
        parentModel[parentModelProp].replace(array);
        return null;
    }
    change.newValue = mobx.observable.array(array);
    return change;
}

var Initializers = {
    object: function object$$1() {
        return mobx.observable.object({});
    },
    array: createCollection
};

var TypeInitializers = {
    hasMany: createCollection
};

function getInitializer(type, options, propName) {
    var fn = TypeInitializers[type] || Initializers[options.type];
    if (!fn) {
        return undefined;
    }
    return function () {
        // eslint-disable-line func-names
        return fn(options, this, propName);
    };
}

function installModelInterceptor(interceptingFn, _ref2) {
    var target = _ref2.target,
        property = _ref2.property;

    var _getSchema = getSchema(target.constructor),
        properties = _getSchema.properties;

    addLazyInitializer(target, function (model) {
        var schemaProps = properties.get(property).options;
        mobx.intercept(model, property, function (change) {
            return interceptingFn(change, {
                inverseOf: schemaProps.inverseOf,
                modelClass: schemaProps.model,
                parentModel: model,
                parentModelProp: property,
                defaultAttributes: schemaProps.defaults
            });
        });
    });
}

function addAttribute(type, setAttributeFn, target, property, descriptor) {
    var options = arguments.length > 5 && arguments[5] !== undefined ? arguments[5] : {};

    getSchema(target.constructor).properties.set(property, { name: property, type: type, options: options });
    var initializer = getInitializer(type, options, property);
    if (initializer) {
        descriptor.initializer = initializer;
    }
    var definition = mobx.observable(target, property, descriptor);
    if ('belongsTo' === type && !options.model) {
        options.model = property;
    }
    if (options.model) {
        installModelInterceptor(setAttributeFn, {
            target: target, property: property, descriptor: descriptor, model: options.model
        });
    }
    var setter = definition.set;

    definition.set = function () {
        attachInterceptors(this);

        for (var _len = arguments.length, args = Array(_len), _key = 0; _key < _len; _key++) {
            args[_key] = arguments[_key];
        }

        return setter.call.apply(setter, [this].concat(args));
    };

    return definition;
}

function buildAttributeDecorator(type, args, setAttributeFn) {
    if ('object' === _typeof(args[0]) && 1 === args.length) {
        return function (target, property, descriptor) {
            return addAttribute(type, setAttributeFn, target, property, descriptor, args[0]);
        };
    }
    return addAttribute.apply(undefined, [type, setAttributeFn].concat(toConsumableArray(args), [{}]));
}

var hasMany = function hasMany() {
    for (var _len2 = arguments.length, args = Array(_len2), _key2 = 0; _key2 < _len2; _key2++) {
        args[_key2] = arguments[_key2];
    }

    return buildAttributeDecorator('hasMany', args, onHasManySet);
};
var belongsTo = function belongsTo() {
    for (var _len3 = arguments.length, args = Array(_len3), _key3 = 0; _key3 < _len3; _key3++) {
        args[_key3] = arguments[_key3];
    }

    return buildAttributeDecorator('belongsTo', args, onPropertySet);
};
var field = function field() {
    for (var _len4 = arguments.length, args = Array(_len4), _key4 = 0; _key4 < _len4; _key4++) {
        args[_key4] = arguments[_key4];
    }

    return buildAttributeDecorator('field', args, onPropertySet);
};
var session = function session() {
    for (var _len5 = arguments.length, args = Array(_len5), _key5 = 0; _key5 < _len5; _key5++) {
        args[_key5] = arguments[_key5];
    }

    return buildAttributeDecorator('session', args, onPropertySet);
};
var identifier$1 = function identifier$$1() {
    for (var _len6 = arguments.length, args = Array(_len6), _key6 = 0; _key6 < _len6; _key6++) {
        args[_key6] = arguments[_key6];
    }

    return buildAttributeDecorator('identifier', args, onPropertySet);
};

exports.registerModel = registerModel;
exports.findModel = findModel;
exports.rememberModelUsing = rememberModelUsing;
exports.lookupModelUsing = lookupModelUsing;
exports.getSchema = getSchema;
exports.registerCustomType = registerCustomType;
exports.identifiedBy = identifiedBy;
exports.unresolvedAssociations = unresolvedAssociations;
exports.field = field;
exports.session = session;
exports.belongsTo = belongsTo;
exports.hasMany = hasMany;
exports.identifier = identifier$1;
exports.createCollection = createCollection;

Object.defineProperty(exports, '__esModule', { value: true });

})));
//# sourceMappingURL=build.full.js.map
